// ----------------------------------------------------------------------
//
// input -- tests of the correctness input to various classes 
//
// Usage:
//   input > input.log 
//
// Currently tested:
//
// M. Fischler	    11/19/99	First version:
//				Hep3Vector, HepEulerAngles, HepAxisAngle
// 
// ----------------------------------------------------------------------

#include "CLHEP/Vector/AxisAngle.h"
#include "CLHEP/Vector/EulerAngles.h"
#include "CLHEP/Vector/ThreeVector.h"
#include "CLHEP/Vector/TwoVector.h"
#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/config/TemplateFunctions.h"
#include "CLHEP/config/iostream.h"
#include "CLHEP/config/fstream.h"

#include <vector>

#ifdef HEP_USE_STD
using HepStd::cin;
using HepStd::cout;
using HepStd::cerr;
using HepStd::endl;
//#ifndef _WIN32
//using HepStd::exp;
//#endif
#endif


// **********************
//
// SECTION I - General tools for the various tests
//
// **********************

// **********************
//
// SECTION II - Preparation 
//
// **********************

void prepare (HepStd::vector<Hep3Vector>  & vv, 
	      HepStd::vector<HepEulerAngles> & ee, 
	      HepStd::vector<HepAxisAngle>  & aa,
	      HepStd::vector<Hep2Vector> & tt ) {
  vv.push_back (Hep3Vector ( .25, -.5, 2.5 ) );
  vv.push_back (Hep3Vector ( 1.0E15, 1<<25, -7.5E-10 ) );
  ee.push_back (HepEulerAngles ( .25, -.5, 1.0 )  );
  ee.push_back (HepEulerAngles ( .6, -.7, .8 )  );
  aa.push_back (HepAxisAngle (Hep3Vector(3, 4, 12), .75) );
  tt.push_back (Hep2Vector ( .25, -.5 ) );
  tt.push_back (Hep2Vector ( 1.0E15, 1<<25 ) );
  return;
}

// **********************
//
// SECTION II - Output
//
// **********************


void output  (HepStd::vector<Hep3Vector>  & vv, 
	      HepStd::vector<HepEulerAngles> & ee, 
	      HepStd::vector<HepAxisAngle>  & aa, 
	      HepStd::vector<Hep2Vector> & tt,
	      HepStd::ofstream & f ) {
  unsigned int i;
  for ( i=0; i< vv.size(); ++i ) {
    f << vv[i] << "\n";	// Standard format
    f << vv[i].x() << " " << vv[i].y() << " " << vv[i].z() << "\n"; // old in
    f << vv[i].x() << "," << vv[i].y() << " " << vv[i].z() << "\n"; 
    f << vv[i].x() << " , " << vv[i].y() << " , " << vv[i].z() << "\n"; 
    f << "(" << vv[i].x() << " " << vv[i].y() << " " << vv[i].z() << ")\n"; 
  }
  f << "\n";
  for ( i=0; i< ee.size(); ++i ) {
    f << ee[i] << "\n";	// Standard format
    f << ee[i].phi() << " " << ee[i].theta() << " " << ee[i].psi() << "\n"; // old in
    f << ee[i].phi() << "," << ee[i].theta() << " " << ee[i].psi() << "\n"; 
    f << ee[i].phi() << " , " << ee[i].theta() << " , " << ee[i].psi() << "\n"; 
    f << "(" << ee[i].phi() << " " << ee[i].theta() << " " << ee[i].psi() << ")\n"; 
  }
  f << "\n";
  for ( i=0; i< aa.size(); ++i ) {
    f << aa[i] << "\n";	// Standard format
    f << aa[i].axis().x() << " " << aa[i].axis().y() << " " << aa[i].axis().z() 
              << " " << aa[i].delta() << "\n"; 	// old in
  }
  f << "\n";
  for ( i=0; i< tt.size(); ++i ) {
    f << tt[i] << "\n";	// Standard format
    f << tt[i].x() << " " << tt[i].y() << "\n"; 
    f << tt[i].x() << "," << tt[i].y() << "\n"; 
    f << tt[i].x() << " , " << tt[i].y() << "\n"; 
    f << "(" << tt[i].x() << " " << tt[i].y() << ")\n"; 
  }
}

// **********************************
//
// SECTION III - Validation of Input
//
// *********************************


int input   ( HepStd::vector<Hep3Vector>  & vv, 
	      HepStd::vector<HepEulerAngles> & ee, 
	      HepStd::vector<HepAxisAngle>  & aa, 
	      HepStd::vector<Hep2Vector> & tt,
	      HepStd::ifstream & f ) {
  unsigned int i, k;
  Hep3Vector v;
  HepEulerAngles e;
  HepAxisAngle a;
  Hep2Vector t;

  cout << "\nHep3Vector:\n";
  for ( i=0; i< vv.size(); ++i ) {
    for (k=1; k <= 5; ++k) {
      f >> v;
      if ( v.isNear(vv[i], 1.0E-5) ) {
        cout << ".";
      } else {
        cout << "3Vector input failed at i = " << i << " k = " << k << "\n";
	cout << "v is " << v << " but vv[i] is " << vv[i] << "\n";
        return 100*i+k;
      }
    }
  }

  cout << "\nHepEulerAngles:\n";
  for ( i=0; i< ee.size(); ++i ) {
    for (k=1; k <= 5; ++k) {
      f >> e;
      if ( e.isNear(ee[i], 1.0E-5) ) {
        cout << ".";
      } else {
        cout << "EulerAngle input failed at i = " << i << " k = " << k << "\n";
	cout << "e is " << e << " but ee[i] is " << ee[i] << "\n";
        return 1000+100*i+k;
      }
    }
  }

  cout << "\nHepAxisAngle:\n";
  for ( i=0; i< aa.size(); ++i ) {
    for (k=1; k <= 2; ++k) {
      f >> a;
      if ( a.isNear(aa[i], 1.0E-5) ) {
        cout << ".";
      } else {
        cout << "AxisAngle input failed at i = " << i << " k = " << k << "\n";
	cout << "a is \n" << a << " but aa[i] is \n" << aa[i] << "\n";
        return 2000+100*i+k;
      }
    }
  }

  cout << "\nHep2Vector:\n";
  for ( i=0; i< tt.size(); ++i ) {
    for (k=1; k <= 5; ++k) {
      f >> t;
      if ( t.isNear(tt[i], 1.0E-5) ) {
        cout << ".";
      } else {
        cout << "2Vector input failed at i = " << i << " k = " << k << "\n";
	cout << "t is " << t << " but tt[i] is " << tt[i] << "\n";
        return 3000+100*i+k;
      }
    }
  }
  cout << "\n";
  return 0;
}




// **********************
//
// SECTION IV - Main
//
// **********************

int main() {

  int mask = 0;
  
  cout << "Enter 0 for output or 1 for input: ";
  int io;
  cin >> io;
  HepStd::vector<Hep3Vector> vv;
  HepStd::vector<HepEulerAngles> ee;
  HepStd::vector<HepAxisAngle> aa;
  HepStd::vector<Hep2Vector> tt;
  prepare (vv, ee, aa, tt);
  if (io) {
    HepStd::ifstream f ("inputTest.out");
    if (!f) {
      cout << "Could not open inputTest.out\n";
      return -1;
    }
    mask |= input(vv, ee, aa, tt, f);
  } else {
    HepStd::ofstream f ("inputTest.out");
    output(vv, ee, aa, tt, f);
  }
  if (mask) {
    cout << "Problems were found - mask is " << mask << "\n";
  }
  return mask;
}

